"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const _ = require("lodash");
const fs = require("fs");
const glob = require("glob");
const tyr_1 = require("./tyr");
require("./type/array");
require("./type/bitmask");
require("./type/boolean");
require("./type/date");
require("./type/daterange");
require("./type/datetime");
require("./type/double");
require("./type/duration");
require("./type/email");
require("./type/image");
require("./type/integer");
require("./type/link");
require("./type/mongoId");
require("./type/object");
require("./type/password");
require("./type/string");
require("./type/text");
require("./type/time");
require("./type/timezone");
require("./type/uid");
require("./type/units");
require("./type/url");
const type_1 = require("./core/type");
require("./catalog");
require("./core/appError");
require("./core/component");
require("./core/collection");
require("./core/event");
require("./core/field");
require("./core/function");
require("./core/index");
require("./core/instance");
require("./core/path");
require("./core/subscription");
require("./core/userError");
require("./core/query");
require("./core/update");
const service_1 = require("./service");
require("./counter");
require("./fake");
require("./migrator");
require("./timer");
require("./media-type");
require("./type/s3");
require("./log/log");
require("./grid/grid");
require("./grid/import");
require("./grid/export");
require("./grid/csv");
require("./grid/excel");
require("./diff/diff");
require("./secure/secure");
require("./secure/secureError");
require("./ui/component-config");
require("./page/markup");
require("./page/page");
require("./unit/unitSystem");
require("./unit/unitFactor");
require("./unit/unitType");
require("./unit/unit");
require("./unit/units");
const express_1 = require("./express");
require("./schema");
const options = tyr_1.default.options;
/*
   TODO:

   * server inherits client objects

   * validation (server vs. client)

   * authorized methods + filtering attributes to the client

   * offline support

   * store cache on client

   * link ownership

   * how do we share code between server and client from the same source .js file ?

   * database integrity analysis (find orphans, bad links, etc.)

   * pre-calc aggregation

  Collection Schema BNF:

  <field def>: {
    is:    <string, a field type>,
    as:    <string, a label>,
    help:  <string, end-user notes>,
    notes: <string, developer notes>,
    link:  <string, a collection name ... is is implied for this>
    ...
  }

  <field>:
      <field def>
    | [ <field> ]
    | <field object>

  <field object>:
    { ( <field name>: <field> )* }

  <schema>: {
    name:  <string>,
    id:    <3-character alphanum beginning with a non-hex alphanum character>,
    db:    <mongodb database>, // optional, if not present will default to options.db
    fields: <field object>
  }

*/
const bootstrappedComponents = [];
_.assign(tyr_1.default, {
    version: require('../../package.json').version,
    generateClientLibrary: express_1.generateClientLibrary,
    _configCalled: false,
    async config(opts) {
        if (!opts) {
            return options;
        }
        if (this._configCalled) {
            throw new Error('Tyr.config() called more than once');
        }
        this._configCalled = true;
        tyr_1.default.clear(options);
        _.extend(options, opts);
        let { mongoClient, db } = opts;
        if (db && !mongoClient) {
            throw new Error(`if specifying a 'db' option you must also specify a 'mongoClient' option`);
        }
        else if (!db && mongoClient) {
            db = mongoClient.db();
        }
        this.mongoClient = mongoClient;
        if (db) {
            this.db = db;
            tyr_1.default.collections.forEach(collection => {
                if (!collection.db) {
                    const server = collection.server;
                    collection.db = server
                        ? this.servers[server]
                        : db.collection(collection.def.dbName);
                }
            });
        }
        else {
            if (process.env.DEBUG) {
                console.warn('******** no "db" property passed to config, boostraping Tyranid without database! ********');
            }
        }
        if (opts.validate) {
            await this.validate(opts.validate);
        }
        // ensure permission defaults
        const p = (options.permissions = options.permissions || {});
        for (const perm of ['find', 'insert', 'update', 'remove']) {
            if (!p[perm]) {
                p[perm] = perm;
            }
        }
        if (opts.indexes && opts.db) {
            await this.createIndexes();
        }
    },
    _validateCalled: false,
    async validate(opts) {
        if (this._validateCalled) {
            throw new Error('Tyr.validate() called more than once');
        }
        this._validateCalled = true;
        if (opts && opts !== true) {
            function process(dirOpts) {
                const globPattern = dirOpts.glob;
                if (globPattern) {
                    for (const file of glob.sync(globPattern, {})) {
                        require(file);
                    }
                }
                else {
                    if (!dirOpts.dir) {
                        throw new Error('dir not specified in validate option.');
                    }
                    const fileRe = dirOpts.fileMatch
                        ? new RegExp(dirOpts.fileMatch)
                        : undefined;
                    fs.readdirSync(dirOpts.dir)
                        .filter(file => !fileRe || fileRe.test(file))
                        .forEach(file => {
                        const fileName = dirOpts.dir + '/' + file;
                        if (!fs.lstatSync(fileName).isDirectory()) {
                            require(fileName);
                        }
                    });
                }
            }
            if (_.isArray(opts)) {
                opts.forEach(process);
            }
            else {
                process(opts);
            }
        }
        const secure = options.secure;
        if (secure) {
            // TODO:  if options.secure is an array of Secures, set Tyr.secure to a
            //        composite Secure that has the array of options.secure as children
            tyr_1.default.secure = secure;
            tyr_1.default.components.push(secure);
        }
        async function bootstrap(stage) {
            const bootstrapping = tyr_1.default.components.filter(col => col.boot && !_.includes(bootstrappedComponents, col));
            let reasons;
            for (let pass = 1; bootstrapping.length && pass < 100; pass++) {
                reasons = [];
                for (let i = 0; i < bootstrapping.length;) {
                    let thisReasons = await bootstrapping[i].boot(stage, pass);
                    if (thisReasons && !_.isArray(thisReasons)) {
                        thisReasons = [thisReasons];
                    }
                    if (thisReasons && thisReasons.length) {
                        reasons.push(...thisReasons);
                        i++;
                    }
                    else {
                        if (stage === 'post-link') {
                            bootstrappedComponents.push(bootstrapping[i]);
                        }
                        bootstrapping.splice(i, 1);
                    }
                }
            }
            if (bootstrapping.length) {
                throw new Error(`Tyranid could not boot during ${stage} stage after 100 passes.\n\n` +
                    'Deadlocked collections: ' +
                    bootstrapping.map(c => c.def.name).join(', ') +
                    '\n\nReasons:\n' +
                    reasons.map(r => '  ' + r).join('\n'));
            }
        }
        await bootstrap('compile');
        // TODO:  move this to log.boot
        function parseLogLevel(name) {
            const ll = options[name];
            if (_.isString(ll)) {
                options[name] = tyr_1.default.byName.tyrLogLevel.byLabel(ll);
                if (!options[name]) {
                    throw new Error(`Unknown ${name}: "${ll}".`);
                }
            }
        }
        parseLogLevel('logLevel');
        parseLogLevel('clientLogLevel');
        parseLogLevel('consoleLogLevel');
        parseLogLevel('dbLogLevel');
        for (const col of tyr_1.default.collections) {
            col.compile('link');
        }
        await bootstrap('link');
        await bootstrap('post-link');
        for (const col of tyr_1.default.collections) {
            service_1.instrumentServerServices(col);
        }
        const migration = options.migration;
        if (migration && migration.migrate && !tyr_1.default.isJobWorker()) {
            await tyr_1.default.migrate();
        }
    },
    /**
     * Mostly just used by tests, not rigorous.
     * @private
     */
    forget(collectionId) {
        const col = tyr_1.default.byId[collectionId];
        if (col) {
            _.remove(tyr_1.default.collections, col => col.id === collectionId);
            _.remove(bootstrappedComponents, comp => comp.id === collectionId);
            delete tyr_1.default.byId[collectionId];
            delete type_1.default.byName[col.def.name];
        }
    },
});
if (global.__TyranidGlobal) {
    throw new Error(`Multiple versions of tyranid are being required, only one global tyranid can exist! ` +
        `Tried to create tyranid version = ${tyr_1.default.version} but ` +
        `global tyranid version = ${global.__TyranidGlobal.version} exists!`);
}
exports.computed = true;
exports.readonly = true;
exports.required = true;
exports.labelField = true;
exports.labelImageField = true;
exports.orderField = true;
global.__TyranidGlobal = exports.Tyr = tyr_1.default;
exports.default = tyr_1.default;
//# sourceMappingURL=root.js.map