"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const faker = require("faker");
const tyranid_1 = require("tyranid");
/**
 * sanitize a database based on tyranid schema
 *
 * @param tyr
 * @param opts
 */
async function sanitize(iOpts) {
    const opts = iOpts || tyranid_1.Tyr.options.sanitize || {};
    const { outDbName = tyranid_1.Tyr.db.databaseName + '____sanitized', batchSize = 200, seed, } = opts;
    const { log, error } = createLogger(opts);
    const admin = await tyranid_1.Tyr.db.admin();
    const existingDbs = await admin.listDatabases();
    if (existingDbs.databases.find((d) => d.name === outDbName)) {
        return error(`Database named ${outDbName} already exists.`);
    }
    if (seed)
        faker.seed(seed);
    log(`Creating sanitized version of ${tyranid_1.Tyr.db.databaseName} in new database: ${outDbName}`);
    const outDb = tyranid_1.Tyr.mongoClient.db(outDbName);
    /**
     * - for each collection
     * - create cursor of all docs
     * - for each doc, scramble text fields with 'sanitize: true' fields
     * - insert into new db with given name...
     */
    const sanitizeCollection = async (collection) => {
        let skip = 0;
        let total = 0;
        const sanitizer = createDocumentSanitizer(collection.def, opts);
        const outCollection = outDb.collection(collection.def.dbName);
        const next = async () => {
            const docs = await collection.findAll({
                query: {},
                skip,
                limit: batchSize,
            });
            skip += batchSize;
            return docs;
        };
        let docs = await next();
        while (docs.length) {
            total += docs.length;
            await outCollection.insertMany(docs.map(sanitizer));
            log(`sanitized ${total} docs from collection ${collection.def.name}`);
            docs = await next();
        }
    };
    const collections = tyranid_1.Tyr.collections.filter(c => !c.def.internal);
    if (opts.serial) {
        for (const collection of collections) {
            await sanitizeCollection(collection);
        }
    }
    else {
        await Promise.all(collections.map(sanitizeCollection));
    }
}
exports.sanitize = sanitize;
tyranid_1.Tyr.sanitize = sanitize;
/**
 * sanitize a single document given its schema
 *
 * @param def
 * @param doc
 */
function createDocumentSanitizer(def, opts) {
    const { autoSanitize } = opts;
    const { warn } = createLogger(opts);
    const skip = (state, value) => {
        warn(`skipping field ${state.path} on document id=${state.$id} (${def.name}) as it doesn't match the schema`);
        return value;
    };
    /**
     * document walker
     *
     * @param doc
     * @param fields
     */
    const walk = (value, fieldDef, state) => {
        const { is } = fieldDef;
        if (!is)
            return value;
        switch (is) {
            case 'array': {
                const { of } = fieldDef;
                if (!of || !value)
                    return value;
                if (!Array.isArray(value))
                    return skip(state, []);
                return value.map((i) => walk(i, typeof of === 'string' ? { is: of } : of, extendPaths('[]', state)));
            }
            case 'object': {
                const fields = fieldDef.fields;
                if (!fields || !value)
                    return value;
                if (typeof value !== 'object')
                    return skip(state, {});
                const out = {};
                for (const field in fields) {
                    const innerFieldDef = fields[field].def;
                    if (!innerFieldDef) {
                        out[field] = value[field];
                    }
                    else {
                        out[field] = walk(value[field], innerFieldDef, extendPaths(field, state));
                    }
                }
                return out;
            }
            default: {
                const sanitizeConfig = fieldDef.sanitize;
                if (sanitizeConfig) {
                    return getSanitizedValue(value, sanitizeConfig);
                }
                if (autoSanitize && is === 'string') {
                    return getSanitizedValue(value, true);
                }
                return value;
            }
        }
    };
    /**
     * return sanitizer
     */
    return (doc) => {
        const out = {};
        const state = extendPaths(def.name, { path: '', $id: doc.$id.toString() });
        for (const field in def.fields) {
            out[field] = walk(doc[field], def.fields[field].def, extendPaths(field, state));
        }
        return out;
    };
}
function getSanitizedValue(defaultValue, sanitizeConfig) {
    if (!sanitizeConfig)
        return defaultValue;
    switch (sanitizeConfig) {
        case 'name':
            return faker.name.findName();
        case 'email':
            return faker.internet.email();
        case 'lorem':
            return faker.lorem.sentences();
        default:
            return faker.lorem.text();
    }
}
function createLogger(opts) {
    const { verbose } = opts;
    const format = (str) => `tyranid-sanitize: ${str}`;
    const log = (str) => verbose && console.log(format(str));
    const warn = (str) => verbose && console.warn(format(str));
    const error = (str) => {
        throw new Error(str);
    };
    return {
        format,
        log,
        error,
        warn,
    };
}
function extendPaths(path, state) {
    return {
        ...state,
        path: `${state.path}.${path}`,
    };
}
//# sourceMappingURL=sanitize.js.map